package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.call_log_output.LogOutput;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.UploadProgressTracker;
import be.iminds.ilabt.jfed.experiment.setup.config.ESpecLogStorageDetails;
import be.iminds.ilabt.jfed.testing.base.ApiTestResult;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * ESpecLogToTestMethodResult: listens to an ESpec, and stores the log in an ApiTestResult.ApiTestMethodResult
 */
public class ESpecLogToTestMethodResult  implements ESpecLogListener {
    private Instant start;
    private Instant stop;

    @Nonnull private final List<ESpecLogStorageDetails> logStorageList;

    @Nonnull private final String name;
    @Nonnull private String uniqueEspecId = "";
    @Nonnull private final String description;
    @Nonnull private final List<LogOutput.LogEntry> logEntries = new ArrayList<>();
    @Nonnull private LogOutput.TestResultState finalState = LogOutput.TestResultState.SUCCESS;

    @Nonnull private List<ApiTestResult.ApiTestMethodResult.FedmonResultExtra> fedmonResultExtras = new ArrayList<>();

    public ESpecLogToTestMethodResult(@Nonnull String name, @Nonnull String description,
                                      @Nullable List<ESpecLogStorageDetails> logStorageList) {
        this.name = name;
        this.description = description;
        this.logStorageList = logStorageList == null ? Collections.emptyList() : logStorageList;
    }

    /** ID used when storing logs of different ESpecs in one place. (as GuiLogicTest can do) */
    public void setUniqueEspecId(@Nonnull String uniqueEspecId) {
        this.uniqueEspecId = uniqueEspecId;
    }

    private void logInstant() {
        if (start == null) {
            start = Instant.now();
        }
        stop = Instant.now();
    }

    public ApiTestResult.ApiTestMethodResult getLogAsTestMethodResult() {
        long durationMs = start == null || stop == null ?
                        0L : stop.toEpochMilli() - start.toEpochMilli();
        ApiTestResult.ApiTestMethodResult res = new ApiTestResult.ApiTestMethodResult(
                name,
                description,
                Collections.emptyList(), //soft deps
                Collections.emptyList(), //hard deps
                Collections.emptyList(), //embedded
                logEntries,
                Collections.emptyList(), //apiCallDetails
                null, //exception
                finalState == null ? LogOutput.TestResultState.SKIPPED : finalState,
                durationMs,
                start == null ? 0L : start.toEpochMilli(),
                stop == null ? 0L : stop.toEpochMilli()
        );
        return res;
    }


    public void errorNonFatal(@Nullable String val) {
        finalState = ApiTestResult.TestResultState.FAILED;
        logEntries.add(new LogOutput.LogEntry(LogOutput.LogLineType.ERROR, val));
    }

    public void note(String val) {
        logEntries.add(new LogOutput.LogEntry(LogOutput.LogLineType.NOTE, val));
    }

    public void addExtraResult(@Nonnull ApiTestResult.ApiTestMethodResult.FedmonResultExtra extra) {
        fedmonResultExtras.add(extra);
    }
    public void addExtraResult(@Nonnull ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder extra) {
        addExtraResult(extra.create());
    }
    @Nonnull
    public List<ApiTestResult.ApiTestMethodResult.FedmonResultExtra> getFedmonResultExtras() {
        return Collections.unmodifiableList(fedmonResultExtras);
    }


    @Override
    public void onPreFileLoad() {
        note("ExperimentSpecification: Start loading files.");
    }

    @Override
    public void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize, @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            errorNonFatal("ExperimentSpecification: Failed to load file \""+fileSource.getBasename()+"\"");
        else {
            if (!isFast)
                note("ExperimentSpecification: Successfully loaded file \""+fileSource.getBasename()+"\" ("+byteSize+" byte)");
        }
    }

    @Override
    public void onPostFileLoadAll(boolean success) {
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to load all files.");
        else
            note("ExperimentSpecification: Successfully finished loading files.");
    }

    @Override
    public void onPreRSpec() {
        note("ExperimentSpecification: Start Provisioning RSpec.");
    }

    @Override
    public void onRequestRSpecKnown(@Nonnull RspecSpec rspecSpec, @Nonnull String requestRspec) {
        note("ExperimentSpecification: onRequestRSpecKnown length="+requestRspec.length());
    }

    @Override
    public void onPostRSpec(@Nonnull RspecSpec rspec, @Nonnull ESpecStepStatus status) {

    }

    @Override
    public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to Provision RSpec");
        else
            note("ExperimentSpecification: Succesfully Provisioned RSpec with nodes "+allNodeClientIds.stream().collect(Collectors.joining()));
    }

    @Override
    public void onPreDir() {
        logInstant();
        note("ExperimentSpecification: Start dir setup.");
    }

    @Override
    public void onPreDirNode(@Nonnull String nodeClientId) {
        logInstant();
    }

    @Override
    public void onPreDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId) {
        logInstant();
    }

    @Override
    public void onPostDir(long logEventId, @Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
        if (status.isFailure())
            errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to setup dir at \""+fullPath+"\"");
        else
            note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully setup dir at \""+fullPath+"\"");
    }

    @Override
    public void onPostDirNode(@Nonnull String nodeClientId, boolean success) {
        logInstant();
    }

    @Override
    public void onPostDirAll(boolean success) {
        logInstant();
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to setup dirs");
        else
            note("ExperimentSpecification: Succesfully setup dirs");
    }

    @Override
    public void onPreUpload() {
        logInstant();
        note("ExperimentSpecification: Start upload(s).");
    }

    @Override
    public void onPreUploadNode(@Nonnull String nodeClientId) {
        logInstant();
    }

    @Override
    public void onPreUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, @Nonnull UploadProgressTracker uploadProgressTracker) {
        logInstant();
    }

    @Override
    public void onPostUpload(long logEventId, @Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath,
                             @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
        logInstant();
        if (status.isFailure())
            errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to upload file \""+uploadSpec.getDesc()+"\" at \""+fullPath+"\"");
        else
            note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully uploaded file \"" + uploadSpec.getDesc() + "\" at \""+fullPath+"\"");
    }

    @Override
    public void onPostUploadNode(@Nonnull String nodeClientId, boolean success) {
        logInstant();
    }

    @Override
    public void onPostUploadAll(boolean success) {
        logInstant();
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to upload at least one file to one node");
        else
            note("ExperimentSpecification: Succesfully uploaded files");
    }

    @Override
    public void onPreExecute() {
        logInstant();
        note("ExperimentSpecification: Start execute(s).");
    }

    private final Map<Long, LimitedLiveLog> tmpLogs = new HashMap<>();
    @Override
    public void onPreExecute(long logEventId,
                             @Nonnull ExecuteSpec executeSpec,
                             @Nonnull String fullExecutablePath,
                             @Nullable String fullLogPath,
                             @Nonnull String nodeClientId,
                             @Nonnull LimitedLiveLog currentLog) {
        logInstant();
        tmpLogs.put(logEventId, currentLog);
    }

    @Override
    public void onPostExecute(long logEventId,
                              @Nonnull ExecuteSpec executeSpec,
                              @Nonnull String fullPath,
                              @Nullable String remoteLogPath,
                              @Nonnull String nodeClientId,
                              int exitStatus, @Nonnull ESpecStepStatus status) {
        logInstant();
        if (status.isFailure()) {
            errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to execute script \"" + executeSpec.getDesc() + "\" at \"" + fullPath + "\" exitStatus=" + exitStatus);
        } else {
            note("ExperimentSpecification: (on \"" + nodeClientId + "\") Successfully executed script \"" + executeSpec.getDesc() + "\" at \"" + fullPath + "\" exitStatus=" + exitStatus);
        }

        if (status.isFailure() || mustKeepLog(executeSpec, nodeClientId)) {
            keepLog(logEventId, executeSpec, nodeClientId, fullPath, remoteLogPath, tmpLogs.get(logEventId));
        }

        tmpLogs.remove(logEventId);
    }

    @Override
    public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {
        logInstant();
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to execute script \"" + executeSpec.getDesc() + "\" on at least one node");
        else
            note("ExperimentSpecification: Succesfully executed script \"" + executeSpec.getDesc() + "\" on all nodes");
    }

    @Override
    public void onPostExecuteAll(boolean success) {
        logInstant();
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to execute at least one script on at least one node");
        else
            note("ExperimentSpecification: Succesfully executed all scripts");
        tmpLogs.clear();
    }

    @Override
    public void onPreAnsible() {
        logInstant();
        note("ExperimentSpecification: Start ansible(s).");
    }

    @Override
    public void onPreAnsiblePlaybook(long logEventId, @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                     @Nonnull String fullPlaybookPath, @Nullable String fullLogPath,
                                     @Nonnull String nodeClientId, @Nonnull LimitedLiveLog currentLog) {
        logInstant();
    }

    @Override
    public void onPostAnsiblePlaybook(long logEventId,
                                      @Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec,
                                      @Nonnull String fullPath,
                                      @Nonnull String remoteLogPath,
                                      @Nonnull String nodeClientId, @Nonnull ESpecStepStatus status) {
        logInstant();
        if (status.isFailure())
            errorNonFatal("ExperimentSpecification: (on \"" + nodeClientId + "\") Failed to execute ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
        else
            note("ExperimentSpecification: (on \"" + nodeClientId + "\") Succesfully executed ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
    }

    @Override
    public void onPostAnsibleAll(boolean success) {
        logInstant();
        if (!success)
            errorNonFatal("ExperimentSpecification: Failed to ansible at least one script on at least one node");
        else
            note("ExperimentSpecification: Succesfully ansibled all scripts");
    }


    @Nullable
    private ESpecLogStorageDetails findLogStorageDetails(
            @Nonnull ExecuteSpec executeSpec,
            @Nonnull String nodeClientId) {
        for (ESpecLogStorageDetails logStorage: logStorageList) {
            if (logStorage.getPath() != null &&
                    executeSpec.getPath() != null &&
                    !executeSpec.getPath().equals(logStorage.getPath()))
                continue;
            if (logStorage.getNodeClientId() != null &&
                    !nodeClientId.equals(logStorage.getNodeClientId()))
                continue;
            return logStorage;
        }
        return null;
    }

    private boolean mustKeepLog(@Nonnull ExecuteSpec executeSpec,
                                @Nonnull String nodeClientId) {
        ESpecLogStorageDetails logStorage =
                findLogStorageDetails(executeSpec, nodeClientId);
        if (logStorage == null)
            return false;
        return logStorage.isStoreInFile() || logStorage.isStoreInLog();
    }

    private void keepLog(long logEventId,
                         @Nonnull ExecuteSpec executeSpec,
                         @Nonnull String nodeClientId,
                         @Nonnull String fullExecutablePath,
                         @Nullable String fullLogPath,
                         @Nonnull LimitedLiveLog currentLog) {
        String useablePath = executeSpec.getPath();
        if (useablePath == null) {
            useablePath = fullLogPath == null? null: fullLogPath.replaceAll(".*/", "");
        }
        if (useablePath == null) {
            useablePath = fullExecutablePath.replaceAll(".*/", "");
        }
        if (useablePath == null) {
            useablePath = "unknown-"+logEventId;
        }

        ESpecLogStorageDetails logStorage =
                findLogStorageDetails(executeSpec, nodeClientId);
        boolean addSeperate = logStorage != null && logStorage.isStoreInFile();
        boolean addInline = logStorage != null && logStorage.isStoreInLog();

        if (!addSeperate && !addInline) {
            //probably an ESpec execute error yhat is now causing the log to be stored
            //while the log wasn't requested to be stored. We will store it in a file in that case.
            addSeperate = true;
        }

        if (addSeperate) {
            addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("log"+uniqueEspecId+"-" + useablePath + "-" + nodeClientId)
                            .setValue(currentLog.getText())
                            .setLarge(true/*currentLog.getReceivedSize() > 100*/)
                            .setHttpMediaTypeToXml()
            );
            logEntries.add(new LogOutput.LogEntry(LogOutput.LogLineType.NOTE,
                    "Log content for "+useablePath+" @ "+nodeClientId+" has been stored. " +
                            "See \""+"log-" + useablePath + "-" + nodeClientId+"\""));
        }
        if (addInline) {
            String logText = currentLog.getText();
            logEntries.add(new LogOutput.LogEntry(LogOutput.LogLineType.NOTE,
                    "Log content for "+useablePath+" @ "+nodeClientId+
                            " ("+logText.length()+" chars):\n\n"+
                            logText, true));
        }
    }
}
